#import <Cocoa/Cocoa.h>
#import <Quartz/Quartz.h>


enum EMode
{
	NUFLI_MODE = 0,
	MUIFLI_MODE = 1
};


enum EInputType
{
	BMP_INPUTTYPE = 0,
	IFLI_INPUTTYPE = 1,
	DRL_INPUTTYPE = 2,
};


enum EPalette
{
	PEPTO_PALETTE = 0,
	DEEKAY_PALETTE = 1,
	TRUECOLOR_PREPARE = 2,
};


enum EInterlaceMode
{
	RASTERED_INTERLACE = 0,
	BLENDED_INTERLACE = 1,
};


@interface CustomProgressIndicator : NSView
{
	BOOL isAnimating;
	float phase;
	float alphaValue;
}

- (void) animate;
- (void) startAnimation:(id)sender;
- (void) stopAnimation:(id)sender;

@end


@interface AppDelegate : NSObject
{
	NSString* mufflonExecutablePath;
	
	EMode activeMode;
	EInputType inputType;
	EPalette sourcePalette;
	EPalette destinationPalette;
	EInterlaceMode nuFliInterlaceMode;
	EInterlaceMode muiFliInterlaceMode;
	
	NSString* inputPath;
	NSString* outputPath;
	NSString* nufliPreConvertPath;
	NSString* muifliPreConvertPath;
	
	NSString* baseName;
	
	BOOL saveResults;
	
	BOOL needsPreviewUpdate;
	NSLock* resultProcessingLock;
	
	IBOutlet NSWindow* settingsWindow;
	IBOutlet NSTabView* settingsTabView;
	//IBOutlet NSProgressIndicator* progressIndicator;
	
	IBOutlet NSTextField* nufliInputPathTextField;
	IBOutlet NSTextField* nufliOutputPathTextField;

	IBOutlet NSTextField* muifliInputPathTextField;
	IBOutlet NSTextField* muifliOutputPathTextField;
	
	IBOutlet NSMatrix* nufliSourcePaletteMatrix;
	IBOutlet NSMatrix* muiFliSourcePaletteMatrix;
	IBOutlet NSMatrix* nufliDestinationPaletteMatrix;
	IBOutlet NSMatrix* muiFliDestinationPaletteMatrix;
	IBOutlet NSPopUpButton* imodePopUpButton;
	IBOutlet NSSlider* nufliLuminanceMin;
	IBOutlet NSSlider* nufliLuminanceMax;
	IBOutlet NSSlider* muifliLuminanceMin;
	IBOutlet NSSlider* muifliLuminanceMax;
	IBOutlet NSTextField* nufliLuminanceMinText;
	IBOutlet NSTextField* nufliLuminanceMaxText;
	IBOutlet NSTextField* muifliLuminanceMinText;
	IBOutlet NSTextField* muifliLuminanceMaxText;
	
	IBOutlet NSButton* renderFliBugButton;
	IBOutlet NSButton* multipassButton;
	
	IBOutlet NSButton* seperateInkButton;
	IBOutlet NSButton* seperatePaperButton;
	IBOutlet NSButton* seperateSpriteButton;
	
	IBOutlet NSButton* deflickerButton;
	IBOutlet NSButton* bruteForceButton;
	IBOutlet NSPopUpButton* fliBugColorPopupButton;
	
	IBOutlet NSButton* saveMuiFliResultButton;
	IBOutlet NSButton* saveNuFliResultButton;
	IBOutlet NSButton* savePreConvertButton;
	
	IBOutlet NSImageView* nufliOriginalImageView;
	IBOutlet NSImageView* nufliPreppedImageView;
	IBOutlet NSImageView* nufliResultImageView;
	IBOutlet NSImageView* nufliErrorImageView;

	IBOutlet NSImageView* muifliOriginalImageView;
	IBOutlet NSImageView* muifliPreppedImageView;
	IBOutlet NSImageView* muifliResultImageView;
	IBOutlet NSImageView* muifliErrorImageView;
	IBOutlet NSImageView* muifliFlickerImageView;
	
	IBOutlet NSButton* convertButton;
	IBOutlet NSButton* selectNuFliInputButton;
	IBOutlet NSButton* selectNuFliOutputButton;
	IBOutlet NSButton* selectMuiFliInputButton;
	IBOutlet NSButton* selectMuiFliOutputButton;
	
	IBOutlet NSTabView* nufliLeftTabView;
	IBOutlet NSTabView* muifliLeftTabView;

	IBOutlet CustomProgressIndicator* progressIndicator;
	
	IBOutlet NSPanel* customAboutPanel;
	IBOutlet NSWindow* helpWindow;
	
	IBOutlet NSWindow* folderProgressWindow;
	IBOutlet NSProgressIndicator* folderProgressIndicator;
	IBOutlet NSTextField* folderProgressLabel;
	
	NSString* mufflonNuFliExecutablePath;
	NSString* mufflonMuiFliExecutablePath;
	NSTask* mainMufflonTask;
	BOOL mufflonTaskWasForPreview;
	BOOL mufflonTaskFinished;
	BOOL folderConversionCancelled;
	int currentConversionIndex;
	int currentConversionsFinished;
	
	NSMutableDictionary* taskToOutputPath;
	
	NSMutableArray* batchMufflonTasks;
}


- (IBAction) startConversion:(id)sender;
- (IBAction) startConversionForFolder:(id)sender;
- (IBAction) cancelFolderConversion:(id)sender;

- (IBAction) selectInputFile:(id)sender;
- (IBAction) selectOutputFile:(id)sender;
- (IBAction) savePreConvertImage:(id)sender;
- (void) inputPathChanged:(NSString*)path;
- (void) inputFileDropped:(NSString*)path;

- (void) setActiveMode:(EMode)mode;

- (void) setInputTypeFromExtension:(NSString*)extension;
- (void) runMufflonTask:(NSTask*)task forPreview:(BOOL)preview forBatch:(BOOL)batch;

- (IBAction) changeSourcePalette:(id)sender;
- (IBAction) changeDestinationPalette:(id)sender;
- (IBAction) changeLuminanceMinNuFli:(id)sender;
- (IBAction) changeLuminanceMinMuiFli:(id)sender;
- (IBAction) changeLuminanceMaxNuFli:(id)sender;
- (IBAction) changeLuminanceMaxMuiFli:(id)sender;
- (IBAction) changeInterlaceMode:(id)sender;
- (IBAction) changeFliBug:(id)sender;

- (IBAction) changeMode:(id)sender;

- (IBAction) orderFrontAboutPanel:(id)sender;
- (IBAction) showHelp:(id)sender;

- (void) convertFolderContents:(NSMutableArray*)filesToConvert;
- (void) selectFolderPanelDidEnd:(NSOpenPanel*)openPanel returnCode:(int)returnCode contextInfo:(void*)contextInfo;
- (void) didEndProgressSheet:(NSWindow*)sheet returnCode:(int)returnCode contextInfo:(void*)contextInfo;

- (void) updatePreview;

- (void) updateTimer;

@end


@interface MufflonImageView : NSImageView
{
	
}

@end

