#import "AppDelegate.h"

#include <sys/types.h>
#include <sys/sysctl.h>

static NSString* sMufflonExecutableName = @"mufflon";

static NSString* sTempPath = nil;

static NSArray* supportedFileTypes = [NSArray arrayWithObjects:@"bmp", @"fun", @"gun", @"drl", @"fp2", nil];


@implementation AppDelegate


// ----------------------------------------------------------------------------
- (void) awakeFromNib
// ----------------------------------------------------------------------------
{
	mufflonExecutablePath = [[[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:sMufflonExecutableName] retain];
	
	activeMode = NUFLI_MODE;
	inputType = BMP_INPUTTYPE;
	sourcePalette = PEPTO_PALETTE;
	destinationPalette = PEPTO_PALETTE;
	nuFliInterlaceMode = RASTERED_INTERLACE;
	muiFliInterlaceMode = RASTERED_INTERLACE;
	inputPath = @"";
	outputPath = @"";
	nufliPreConvertPath = nil;
	muifliPreConvertPath = nil;
	baseName = @"";
	saveResults = NO;
	mufflonTaskFinished = YES;
	folderConversionCancelled = NO;
	needsPreviewUpdate = NO;
	
	mainMufflonTask = nil;
	batchMufflonTasks = [[NSMutableArray alloc] init];

	taskToOutputPath = [[NSMutableDictionary alloc] init];
	
	resultProcessingLock = [[NSLock alloc] init];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(inputPathChangedNotification:)
												 name:NSControlTextDidChangeNotification
											   object:nil];
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(mufflonTaskFinished:) name:NSTaskDidTerminateNotification object:nil];	

	FSRef folder;
	OSErr err = FSFindFolder(kUserDomain, kUserSpecificTmpFolderType, false, &folder);
	if (err == noErr)
	{
		CFURLRef url =  CFURLCreateFromFSRef(kCFAllocatorDefault, &folder);
		sTempPath = [(NSURL*)url path];
		sTempPath = [sTempPath stringByAppendingPathComponent:@"Mufflon"];
		
		BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:sTempPath];
		if (!exists)
			[[NSFileManager defaultManager] createDirectoryAtPath:sTempPath attributes:nil];
		
		[sTempPath retain];
	}
	
	[settingsTabView selectTabViewItemAtIndex:activeMode];
	
	NSTimer* timer = [NSTimer scheduledTimerWithTimeInterval:0.05f target:self selector:@selector(updateTimer) userInfo:nil repeats:YES];
	[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSEventTrackingRunLoopMode];
}


// ----------------------------------------------------------------------------
- (IBAction) orderFrontAboutPanel:(id)sender
// ----------------------------------------------------------------------------
{
	[customAboutPanel orderFront:sender];
}


// ----------------------------------------------------------------------------
- (IBAction) showHelp:(id)sender
// ----------------------------------------------------------------------------
{
	if ([helpWindow isVisible])
		[helpWindow orderOut:self];
	else
		[helpWindow orderFront:self];
}


// ----------------------------------------------------------------------------
- (void) updateTimer
// ----------------------------------------------------------------------------
{
	[progressIndicator animate];
}


// ----------------------------------------------------------------------------
- (void) applicationWillTerminate:(NSNotification*)aNotification
// ----------------------------------------------------------------------------
{
	if (mainMufflonTask != nil)
		[mainMufflonTask terminate];
}


// ----------------------------------------------------------------------------
- (NSTextField*) getInputPathTextField
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nufliInputPathTextField;
	else
		return muifliInputPathTextField;
}


// ----------------------------------------------------------------------------
- (NSTextField*) getOutputPathTextField
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nufliOutputPathTextField;
	else
		return muifliOutputPathTextField;
}


// ----------------------------------------------------------------------------
- (NSImageView*) getOriginalImageView
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nufliOriginalImageView;
	else
		return muifliOriginalImageView;
}


// ----------------------------------------------------------------------------
- (NSImageView*) getPreConvertImageView
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nufliPreppedImageView;
	else
		return muifliPreppedImageView;
}


// ----------------------------------------------------------------------------
- (NSImageView*) getResultImageView
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nufliResultImageView;
	else
		return muifliResultImageView;
}


// ----------------------------------------------------------------------------
- (NSImageView*) getErrorImageView
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nufliErrorImageView;
	else
		return muifliErrorImageView;
}


// ----------------------------------------------------------------------------
- (NSImageView*) getFlickerImageView
// ----------------------------------------------------------------------------
{
	if (activeMode == NUFLI_MODE)
		return nil;
	else
		return muifliFlickerImageView;
}


// ----------------------------------------------------------------------------
- (NSString*) getExtensionForMode:(BOOL)preview
// ----------------------------------------------------------------------------
{
	if (preview)
		return @"bmp";
	
	if (activeMode == NUFLI_MODE)
		return @"nuf";
	else
		return @"mui";
}



// ----------------------------------------------------------------------------
- (IBAction) selectInputFile:(id)sender
// ----------------------------------------------------------------------------
{
	NSOpenPanel* openPanel = [NSOpenPanel openPanel];
	
	[openPanel beginSheetForDirectory:nil file:nil types:supportedFileTypes modalForWindow:settingsWindow modalDelegate:self didEndSelector:@selector(didEndOpenFileSheet:returnCode:contextInfo:) contextInfo:self];
}


// ----------------------------------------------------------------------------
- (void)didEndOpenFileSheet:(NSOpenPanel*)openPanel returnCode:(int)returnCode contextInfo:(void*)contextInfo
// ----------------------------------------------------------------------------
{
	if (returnCode == NSOKButton)
	{
        NSArray *filesToOpen = [openPanel filenames];
        NSString* file = [filesToOpen objectAtIndex:0];
		
		[self inputPathChanged:file];
    }	
}


// ----------------------------------------------------------------------------
- (void) setInputTypeFromExtension:(NSString*)extension
// ----------------------------------------------------------------------------
{
	if ([extension caseInsensitiveCompare:@"fun"] == NSOrderedSame || 
		[extension caseInsensitiveCompare:@"gun"] == NSOrderedSame || 
		[extension caseInsensitiveCompare:@"fp2"] == NSOrderedSame)
		inputType = IFLI_INPUTTYPE;
	else if ([extension caseInsensitiveCompare:@"drl"] == NSOrderedSame)
		inputType = DRL_INPUTTYPE;
	else
		inputType = BMP_INPUTTYPE;
	
	[imodePopUpButton setEnabled:inputType != BMP_INPUTTYPE];
}


// ----------------------------------------------------------------------------
- (void) inputPathChanged:(NSString*)path
// ----------------------------------------------------------------------------
{
	[self setInputTypeFromExtension:[path pathExtension]];
	
	if (activeMode == NUFLI_MODE)
	{
		sourcePalette = (EPalette) [nufliSourcePaletteMatrix selectedTag];
		destinationPalette = (EPalette) [nufliDestinationPaletteMatrix selectedTag];

		BOOL prepareEnabled = (inputType == BMP_INPUTTYPE) && (sourcePalette == TRUECOLOR_PREPARE);
		
		[nufliLuminanceMax setEnabled:prepareEnabled];
		[nufliLuminanceMin setEnabled:prepareEnabled];
		[nufliLuminanceMinText setEnabled:prepareEnabled];
		[nufliLuminanceMaxText setEnabled:prepareEnabled];
		[nufliSourcePaletteMatrix setEnabled:(inputType == BMP_INPUTTYPE)];
	}
	else
	{
		sourcePalette = (EPalette) [muiFliSourcePaletteMatrix selectedTag];
		destinationPalette = (EPalette) [muiFliDestinationPaletteMatrix selectedTag];

		BOOL prepareEnabled = (inputType == BMP_INPUTTYPE) && (sourcePalette == TRUECOLOR_PREPARE);
		
		[muifliLuminanceMax setEnabled:prepareEnabled];
		[muifliLuminanceMin setEnabled:prepareEnabled];
		[muifliLuminanceMinText setEnabled:prepareEnabled];
		[muifliLuminanceMaxText setEnabled:prepareEnabled];
		[muiFliSourcePaletteMatrix setEnabled:(inputType == BMP_INPUTTYPE)];
	}
	
	NSImageView* imageView = [self getOriginalImageView];

	if (inputType == BMP_INPUTTYPE)
	{
		if ([[NSFileManager defaultManager] fileExistsAtPath:path])
		{
			NSImage* originalImage = [[NSImage alloc] initByReferencingFile:path];
			//NSLog(@"image: %@\n", originalImage);
			NSBitmapImageRep* imageRep = [[originalImage representations] objectAtIndex:0];
			[originalImage setSize:NSMakeSize([imageRep pixelsWide], [imageRep pixelsHigh])];
			
			//NSLog(@"rep: %@\n", imageRep);
			BOOL validBpp = ([imageRep bitsPerPixel] == 24) || ([imageRep bitsPerPixel] == 32 && ![imageRep hasAlpha]);
			if ([imageRep pixelsWide] != 320 || [imageRep pixelsHigh] != 200 || !validBpp)
			{
				NSAlert* alert = [NSAlert alertWithMessageText:@"Input file needs to be 320x200 in 24bit color!"
												 defaultButton:@"OK"
											   alternateButton:nil
												   otherButton:nil
									 informativeTextWithFormat:@""];
				
				[alert runModal];
				return;
			}
			else
				[imageView setImage:originalImage];
		}
		else
			[imageView setImage:nil];
	}
	else
		[imageView setImage:nil];

	NSTextField* textField = [self getInputPathTextField];
	[textField setStringValue:path];

	textField = [self getOutputPathTextField];
	[textField setStringValue:@""];

	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (void) inputPathChangedNotification:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
	if ([notification object] == nufliInputPathTextField || [notification object] == muifliInputPathTextField)
	{
		NSString* inputPathString = [[notification object] stringValue];
		BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:inputPathString];

		if ([supportedFileTypes containsObject:[inputPathString pathExtension]] && exists)
			[self inputPathChanged:inputPathString];
	}
}


// ----------------------------------------------------------------------------
- (void) inputFileDropped:(NSString*)path
// ----------------------------------------------------------------------------
{
	if ([[path pathExtension] caseInsensitiveCompare:@"bmp"] == NSOrderedSame)
		[self inputPathChanged:[path retain]];
}	



// ----------------------------------------------------------------------------
- (IBAction) selectOutputFile:(id)sender
// ----------------------------------------------------------------------------
{
	NSSavePanel* savePanel = [NSSavePanel savePanel];
	[savePanel setRequiredFileType:[self getExtensionForMode:NO]];
	[savePanel setCanSelectHiddenExtension:NO];
	
	NSTextField* inputPathTextField = [self getInputPathTextField];
	NSString* inputFile = [inputPathTextField stringValue];

	NSString* suggestedFilename = [[[inputFile lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:[self getExtensionForMode:NO]];
	[suggestedFilename retain];
	
	[savePanel beginSheetForDirectory:nil file:suggestedFilename modalForWindow:settingsWindow modalDelegate:self didEndSelector:@selector(savePanelDidEnd:returnCode:contextInfo:) contextInfo:self];
}


// ----------------------------------------------------------------------------
- (void)savePanelDidEnd:(NSSavePanel*)savePanel returnCode:(int)returnCode contextInfo:(void*)contextInfo
// ----------------------------------------------------------------------------
{
	if (returnCode == NSOKButton)
	{
		NSTextField* textField = [self getOutputPathTextField];
		[textField setStringValue:[savePanel filename]];
	}
}


// ----------------------------------------------------------------------------
- (IBAction) savePreConvertImage:(id)sender
// ----------------------------------------------------------------------------
{
	NSString* preConvertPath = activeMode == NUFLI_MODE ? nufliPreConvertPath : muifliPreConvertPath;
	if (preConvertPath == nil)
		return;
	
	NSSavePanel* savePanel = [NSSavePanel savePanel];
	[savePanel setRequiredFileType:@"bmp"];
	[savePanel setCanSelectHiddenExtension:NO];

	NSString* suggestedFilename = [[preConvertPath lastPathComponent] retain];
	
	[savePanel beginSheetForDirectory:nil file:suggestedFilename modalForWindow:settingsWindow modalDelegate:self didEndSelector:@selector(savePreConvertPanelDidEnd:returnCode:contextInfo:) contextInfo:self];
}


// ----------------------------------------------------------------------------
- (void)savePreConvertPanelDidEnd:(NSSavePanel*)savePanel returnCode:(int)returnCode contextInfo:(void*)contextInfo
// ----------------------------------------------------------------------------
{
	if (returnCode == NSOKButton)
	{
		NSString* savePreConvertPath = [savePanel filename];
		
		if (savePreConvertPath != nil)
		{
			NSFileManager* fileManager = [NSFileManager defaultManager];

			NSString* preConvertPath = activeMode == NUFLI_MODE ? nufliPreConvertPath : muifliPreConvertPath;
			
			[fileManager removeFileAtPath:savePreConvertPath handler:nil];
			//NSLog(@"copy: %@ to %@\n", preConvertPath, savePreConvertPath);
			[fileManager copyPath:preConvertPath toPath:savePreConvertPath handler:nil];
		}
	}
}


// ----------------------------------------------------------------------------
- (void) updatePreview
// ----------------------------------------------------------------------------
{
	needsPreviewUpdate = YES;

	if (mainMufflonTask != nil || !mufflonTaskFinished)
	{
		//NSLog(@"mufflon task still running\n");
		return;
	}
	
	mufflonTaskFinished = NO;
	
	NSTextField* inputPathTextField = [self getInputPathTextField];
	inputPath = [inputPathTextField stringValue];

	NSTextField* outputPathTextField = [self getOutputPathTextField];
	outputPath = [outputPathTextField stringValue];
	
	[self runMufflonTask:mainMufflonTask forPreview:YES forBatch:NO];
}


// ----------------------------------------------------------------------------
- (IBAction) changeSourcePalette:(id)sender
// ----------------------------------------------------------------------------
{
	sourcePalette = (EPalette) [sender selectedTag];

	BOOL prepareEnabled = sourcePalette == TRUECOLOR_PREPARE;
	if (activeMode == NUFLI_MODE)
	{
		[nufliLuminanceMax setEnabled:prepareEnabled];
		[nufliLuminanceMin setEnabled:prepareEnabled];
		[nufliLuminanceMinText setEnabled:prepareEnabled];
		[nufliLuminanceMaxText setEnabled:prepareEnabled];
	}
	else
	{
		[muifliLuminanceMax setEnabled:prepareEnabled];
		[muifliLuminanceMin setEnabled:prepareEnabled];
		[muifliLuminanceMinText setEnabled:prepareEnabled];
		[muifliLuminanceMaxText setEnabled:prepareEnabled];
	}
	
	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeDestinationPalette:(id)sender
// ----------------------------------------------------------------------------
{
	destinationPalette = (EPalette) [sender selectedTag];
	
	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeInterlaceMode:(id)sender
// ----------------------------------------------------------------------------
{
	if (activeMode == MUIFLI_MODE)
		muiFliInterlaceMode = (EInterlaceMode) [sender selectedTag];
	else
		nuFliInterlaceMode = (EInterlaceMode) [sender selectedTag];
		
	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeLuminanceMinNuFli:(id)sender
// ----------------------------------------------------------------------------
{
	double minValue = [sender doubleValue];
	double maxValue = [nufliLuminanceMax doubleValue];
	if (minValue < (maxValue - 0.0001f))
	{
		[nufliLuminanceMin setDoubleValue:minValue];
		[nufliLuminanceMinText setDoubleValue:minValue];
	}
	else
	{
		[sender setDoubleValue:maxValue - 0.0001f];
		[nufliLuminanceMin setDoubleValue:maxValue - 0.0001f];
		[nufliLuminanceMinText setDoubleValue:maxValue - 0.0001f];
	}

	//NSLog(@"changed nufli min luminance, min/max: %f/%f\n", minValue, maxValue);
	
	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeLuminanceMinMuiFli:(id)sender
// ----------------------------------------------------------------------------
{
	double minValue = [sender doubleValue];
	double maxValue = [muifliLuminanceMax doubleValue];
	if (minValue < (maxValue - 0.0001f))
	{
		[muifliLuminanceMin setDoubleValue:minValue];
		[muifliLuminanceMinText setDoubleValue:minValue];
	}
	else
	{
		[sender setDoubleValue:maxValue - 0.0001f];
		[muifliLuminanceMin setDoubleValue:maxValue - 0.0001f];
		[muifliLuminanceMinText setDoubleValue:maxValue - 0.0001f];
	}

	//NSLog(@"changed muifli min luminance, min/max: %f/%f\n", minValue, maxValue);

	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeLuminanceMaxNuFli:(id)sender
// ----------------------------------------------------------------------------
{
	double maxValue = [sender doubleValue];
	double minValue = [nufliLuminanceMin doubleValue];
	if (minValue < (maxValue - 0.0001f))
	{
		[nufliLuminanceMax setDoubleValue:maxValue];
		[nufliLuminanceMaxText setDoubleValue:maxValue];
	}
	else
	{
		[sender setDoubleValue:minValue + 0.0001f];
		[nufliLuminanceMax setDoubleValue:minValue + 0.0001f];
		[nufliLuminanceMaxText setDoubleValue:minValue + 0.0001f];
	}

	//NSLog(@"changed nufli max luminance, min/max: %f/%f\n", minValue, maxValue);

	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeLuminanceMaxMuiFli:(id)sender
// ----------------------------------------------------------------------------
{
	double maxValue = [sender doubleValue];
	double minValue = [muifliLuminanceMin doubleValue];
	if (minValue < (maxValue - 0.0001f))
	{
		[muifliLuminanceMax setDoubleValue:maxValue];
		[muifliLuminanceMaxText setDoubleValue:maxValue];
	}
	else
	{
		[sender setDoubleValue:minValue + 0.0001f];
		[muifliLuminanceMax setDoubleValue:minValue + 0.0001f];
		[muifliLuminanceMaxText setDoubleValue:minValue + 0.0001f];
	}

	//NSLog(@"changed muifli max luminance, min/max: %f/%f\n", minValue, maxValue);

	[self updatePreview];
}


// ----------------------------------------------------------------------------
- (IBAction) changeFliBug:(id)sender
// ----------------------------------------------------------------------------
{
	[multipassButton setEnabled:[sender state] == NSOnState];
}


// ----------------------------------------------------------------------------
- (IBAction) changeMode:(id)sender
// ----------------------------------------------------------------------------
{
	[settingsTabView selectTabViewItemAtIndex:[sender tag]];
	[self setActiveMode:(EMode)[sender tag]];
}


// ----------------------------------------------------------------------------
- (void) setActiveMode:(EMode)mode
// ----------------------------------------------------------------------------
{
	activeMode = (EMode) mode;
	
	if (activeMode == NUFLI_MODE)
		[nufliInputPathTextField setStringValue:[muifliInputPathTextField stringValue]];
	else
		[muifliInputPathTextField setStringValue:[nufliInputPathTextField stringValue]];
	
	NSTextField* inputPathTextField = [self getInputPathTextField];
	inputPath = [inputPathTextField stringValue];
	
	if ([inputPath length] != 0)
		[self inputPathChanged:inputPath];
	
	EInterlaceMode interlaceMode = activeMode == MUIFLI_MODE ? muiFliInterlaceMode : nuFliInterlaceMode;
	[imodePopUpButton selectItemWithTag:interlaceMode];
}


// ----------------------------------------------------------------------------
- (IBAction) startConversion:(id)sender
// ----------------------------------------------------------------------------
{
	NSTextField* inputPathTextField = [self getInputPathTextField];
	inputPath = [inputPathTextField stringValue];
	
	NSTextField* outputPathTextField = [self getOutputPathTextField];
	outputPath = [outputPathTextField stringValue];
	
	[self runMufflonTask:mainMufflonTask forPreview:NO forBatch:NO];
}


// ----------------------------------------------------------------------------
- (IBAction) startConversionForFolder:(id)sender
// ----------------------------------------------------------------------------
{
	NSOpenPanel* openPanel = [NSOpenPanel openPanel];
	NSArray* fileTypes = [NSArray arrayWithObject:@""];
	[openPanel setCanChooseDirectories:YES];
	[openPanel setCanChooseFiles:NO];
	[openPanel setAllowsMultipleSelection:NO];
	[openPanel setTitle:@"Select folder to convert"];
	[openPanel setPrompt:@"Choose"];
	
	[openPanel beginSheetForDirectory:nil file:nil types:fileTypes modalForWindow:settingsWindow modalDelegate:self didEndSelector:@selector(selectFolderPanelDidEnd:returnCode:contextInfo:) contextInfo:self];
}



// ----------------------------------------------------------------------------
- (void) selectFolderPanelDidEnd:(NSOpenPanel*)openPanel returnCode:(int)returnCode contextInfo:(void*)contextInfo
// ----------------------------------------------------------------------------
{
	if (returnCode == NSOKButton)
	{
		NSString* folderPath = [openPanel filename];
		[openPanel close];
		
		NSMutableArray* filesToConvert = [[NSMutableArray alloc] init];

		NSDirectoryEnumerator* dirEnumerator = [[NSFileManager defaultManager] enumeratorAtPath:folderPath];
		NSString* file;
		
		while(file = [dirEnumerator nextObject])
		{
			if ([file characterAtIndex:0] == '.')
			{
				[dirEnumerator skipDescendents];
				continue;
			}
			
			NSString* path = [folderPath stringByAppendingPathComponent:file];
			BOOL folder = NO;
			BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:&folder];
			if (!exists)
				continue;
			
			if ([[file pathExtension] caseInsensitiveCompare:@"bmp"] == NSOrderedSame && !folder)
				[filesToConvert addObject:path];
			
			if (folder)
				[dirEnumerator skipDescendents];
		}
		
		[folderProgressIndicator setMaxValue:[filesToConvert count]];
		folderConversionCancelled = NO;
		
		[NSApp beginSheet:folderProgressWindow modalForWindow:settingsWindow modalDelegate:self didEndSelector:@selector(didEndProgressSheet:returnCode:contextInfo:) contextInfo:nil];
		
		[NSThread detachNewThreadSelector:@selector(convertFolderContents:) toTarget:self withObject:filesToConvert];
	}
}



// ----------------------------------------------------------------------------
- (void) convertFolderContents:(NSMutableArray*)filesToConvert
// ----------------------------------------------------------------------------
{
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];	

	NSProcessInfo* processInfo = [NSProcessInfo processInfo];
	int numCores = 1;
	if ([processInfo respondsToSelector:@selector(activeProcessorCount)])
		numCores = (int) [processInfo activeProcessorCount];
	else
	{
		int mib[2] = { CTL_HW, HW_NCPU };
		size_t len = sizeof(int); 
		int rc = sysctl(mib, 2, &numCores, &len, NULL, 0);
		if (rc == -1)
			numCores = 2;
	}

	currentConversionIndex = 0;
	currentConversionsFinished = 0;
	
	mufflonTaskFinished = NO;

	[folderProgressIndicator setDoubleValue:0.0f];

	while (!folderConversionCancelled) 
	{
		[resultProcessingLock lock];
		
		if ([batchMufflonTasks count] < numCores)
		{
			//NSLog(@"Have a free core (used: %d, total: %d)\n", [batchMufflonTasks count], numCores);
			
			inputPath = [filesToConvert objectAtIndex:currentConversionIndex];
			outputPath = nil;
			
			BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:inputPath];
			if (exists)
			{
				NSTask* task = [[NSTask alloc] init];
				[batchMufflonTasks addObject:task];
				
				[self runMufflonTask:task forPreview:NO forBatch:YES];
				
				[folderProgressLabel setStringValue:[NSString stringWithFormat:@"Converting '%@' (%d of %d)", [inputPath lastPathComponent], currentConversionIndex+1, [filesToConvert count]]];
				
				//NSLog(@"Started a task (%@)!\n", inputPath);
			}
			else
				currentConversionsFinished++;

			currentConversionIndex++;
			if (currentConversionIndex == [filesToConvert count])
			{
				//NSLog(@"currentConversionIndex: %d, finishing\n", currentConversionIndex);
				[resultProcessingLock unlock];
				break;
			}
		}

		[resultProcessingLock unlock];

		//NSLog(@"Sleeping thread!\n");

		[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.2f]];
	}

	while (!folderConversionCancelled && currentConversionsFinished != [filesToConvert count])
	{
		//NSLog(@"Waiting for remaining conversions\n");
		[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.2f]];
	}

	if (folderConversionCancelled && [batchMufflonTasks count] > 0)
		[batchMufflonTasks makeObjectsPerformSelector:@selector(terminate)];
	
	[NSApp endSheet:folderProgressWindow];
	[filesToConvert release];
	
	[progressIndicator stopAnimation:self];	

	[pool release];
    [NSThread exit];
}


// ----------------------------------------------------------------------------
- (void) didEndProgressSheet:(NSWindow*)sheet returnCode:(int)returnCode contextInfo:(void*)contextInfo
// ----------------------------------------------------------------------------
{
    [sheet orderOut:self];
	[sheet close];
}


// ----------------------------------------------------------------------------
- (IBAction) cancelFolderConversion:(id)sender
// ----------------------------------------------------------------------------
{
	folderConversionCancelled = YES;
}


// ----------------------------------------------------------------------------
- (void) runMufflonTask:(NSTask*)task forPreview:(BOOL)preview forBatch:(BOOL)batch
// ----------------------------------------------------------------------------
{
	//NSLog(@"Starting mufflon for preview\n");
	
	if ([inputPath length] == 0)
	{
		if (!preview)
		{
			NSAlert* alert = [NSAlert alertWithMessageText:@"You need to select an input file!"
											 defaultButton:@"OK"
										   alternateButton:nil
											   otherButton:nil
								 informativeTextWithFormat:@""];
			
			[alert runModal];
		}
		return;
	}
	
	BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:inputPath];
	if (!exists)
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Input file does not exist!"
										 defaultButton:@"OK"
									   alternateButton:nil
										   otherButton:nil
							 informativeTextWithFormat:@"Please select a valid file and try again."];
		
		[alert runModal];
		
		[[self getInputPathTextField] setStringValue:@""];
		return;
	}
	
	[self setInputTypeFromExtension:[inputPath pathExtension]];
	
	[convertButton setEnabled:NO];
	[nufliInputPathTextField setEnabled:NO];
	[nufliOutputPathTextField setEnabled:NO];
	[muifliInputPathTextField setEnabled:NO];
	[muifliOutputPathTextField setEnabled:NO];
	[selectNuFliInputButton setEnabled:NO];
	[selectNuFliOutputButton setEnabled:NO];
	[selectMuiFliInputButton setEnabled:NO];
	[selectMuiFliOutputButton setEnabled:NO];
	[savePreConvertButton setEnabled:NO];

	if (outputPath == nil || [outputPath length] == 0)
	{
		outputPath = [[[inputPath stringByDeletingPathExtension] stringByAppendingPathExtension:[self getExtensionForMode:preview]] retain];
		if (!preview)
		{
			NSTextField* textField = [self getOutputPathTextField];
			[textField setStringValue:outputPath];
		}
	}
	
	baseName = [[[outputPath lastPathComponent] stringByDeletingPathExtension] retain]; 
	
	NSImageView* imageView = [self getResultImageView];
	if (imageView != nil)
		[imageView setImage:nil];
	
	imageView = [self getErrorImageView];
	if (imageView != nil)
		[imageView setImage:nil];
	
	imageView = [self getFlickerImageView];
	if (imageView != nil)
		[imageView setImage:nil];
	
	NSMutableArray* mufflonArguments = [NSMutableArray arrayWithCapacity:3];
	
	[mufflonArguments addObject:inputPath];

	[mufflonArguments addObject:@"-o"];
	[mufflonArguments addObject:[outputPath lastPathComponent]];
		
	switch (inputType)
	{
		case BMP_INPUTTYPE:
			[mufflonArguments addObject:@"--itype"];
			[mufflonArguments addObject:@"bmp"];
			break;
		case DRL_INPUTTYPE:
			[mufflonArguments addObject:@"--itype"];
			[mufflonArguments addObject:@"drl"];
			break;
		case IFLI_INPUTTYPE:
			[mufflonArguments addObject:@"--itype"];
			[mufflonArguments addObject:@"ifli"];
			break;
	}
	
	bool prepare = false;
	
	if (activeMode == MUIFLI_MODE)
	{
		sourcePalette = (EPalette) [muiFliSourcePaletteMatrix selectedTag];
		destinationPalette = (EPalette) [muiFliDestinationPaletteMatrix selectedTag];

		prepare = sourcePalette == TRUECOLOR_PREPARE;
		
		saveResults = [saveMuiFliResultButton state] == NSOnState;
		
		if (!preview)
		{
			[mufflonArguments addObject:@"--otype"];
			[mufflonArguments addObject:@"muifli"];
		}

		if ([seperateInkButton state] == NSOnState)
			[mufflonArguments addObject:@"--2ink"];

		if ([seperatePaperButton state] == NSOnState)
			[mufflonArguments addObject:@"--2paper"];

		if ([seperateSpriteButton state] == NSOnState)
			[mufflonArguments addObject:@"--2sprite"];

		int fliBugColor = [fliBugColorPopupButton selectedTag];
		[mufflonArguments addObject:@"--bugcol"];
		[mufflonArguments addObject:[NSString stringWithFormat:@"%d", fliBugColor]];

		if ([deflickerButton state] == NSOnState)
			[mufflonArguments addObject:@"-d"];
		
		if ([bruteForceButton state] == NSOnState)
			[mufflonArguments addObject:@"-b"];
	}
	else
	{
		sourcePalette = (EPalette) [nufliSourcePaletteMatrix selectedTag];
		destinationPalette = (EPalette) [nufliDestinationPaletteMatrix selectedTag];
		
		prepare = sourcePalette == TRUECOLOR_PREPARE;
		
		saveResults = [saveNuFliResultButton state] == NSOnState;

		if (!preview)
		{
			[mufflonArguments addObject:@"--otype"];
			[mufflonArguments addObject:@"nufli"];
		}
		
		if ([renderFliBugButton state] == NSOnState)
		{
			[mufflonArguments addObject:@"--flibug"];
		
			if ([multipassButton state] == NSOnState)
				[mufflonArguments addObject:@"--multipass"];
		}
	}
		
	if (inputType != BMP_INPUTTYPE)
		sourcePalette = destinationPalette;
	
	if (sourcePalette == PEPTO_PALETTE)
	{
		[mufflonArguments addObject:@"--src-palette"];
		[mufflonArguments addObject:@"pepto"];
	}
	else if (sourcePalette == DEEKAY_PALETTE)
	{
		[mufflonArguments addObject:@"--src-palette"];
		[mufflonArguments addObject:@"deekay"];
	}
	
	if (destinationPalette == PEPTO_PALETTE)
	{
		[mufflonArguments addObject:@"--dest-palette"];
		[mufflonArguments addObject:@"pepto"];
	}
	else if (destinationPalette == DEEKAY_PALETTE)
	{
		[mufflonArguments addObject:@"--dest-palette"];
		[mufflonArguments addObject:@"deekay"];
	}
	
	if (prepare)
	{
		[mufflonArguments addObject:@"-p"];
		
		float luminanceMin = 0.0f;
		float luminanceMax = 0.0f;
		if (activeMode == NUFLI_MODE)
		{
			luminanceMin = [nufliLuminanceMin floatValue];
			luminanceMax = [nufliLuminanceMax floatValue];
		}
		else
		{
			luminanceMin = [muifliLuminanceMin floatValue];
			luminanceMax = [muifliLuminanceMax floatValue];
		}
		
		[mufflonArguments addObject:@"--min"];
		[mufflonArguments addObject:[NSString stringWithFormat:@"%f", luminanceMin]];
		
		[mufflonArguments addObject:@"--max"];
		[mufflonArguments addObject:[NSString stringWithFormat:@"%f", luminanceMax]];
	}
	
	EInterlaceMode interlaceMode = activeMode == MUIFLI_MODE ? muiFliInterlaceMode : nuFliInterlaceMode;
	[mufflonArguments addObject:@"--imode"];
	if (interlaceMode == BLENDED_INTERLACE)
		[mufflonArguments addObject:@"blended"];
	else
		[mufflonArguments addObject:@"rastered"];
	
	if (preview)
	{
		[mufflonArguments addObject:@"--otype"];
		[mufflonArguments addObject:@"bmp"];
		
		if (activeMode == NUFLI_MODE)
			[mufflonArguments addObject:@"--solid"];
	}
	
	[mufflonArguments addObject:@"--no-truncate"];
	//[mufflonArguments addObject:@"--shutup"];
	
	mufflonTaskWasForPreview = preview;

	if (task == nil)
	{
		task = [[NSTask alloc] init];
		mainMufflonTask = task;
	}
	
	[task setLaunchPath:mufflonExecutablePath];
	[task setCurrentDirectoryPath:sTempPath];
	[task setArguments:mufflonArguments];
	
	[taskToOutputPath setObject:[outputPath copy] forKey:[NSNumber numberWithInt:(int)task]];
	
	[task launch];
	
	//NSLog(@"Launched task: 0x%08x (%@)\n", task, task);
	
	//[progressIndicator setHidden:NO];
	[progressIndicator startAnimation:self];	
}


// ----------------------------------------------------------------------------
- (void) mufflonTaskFinished:(NSNotification*)aNotification
// ----------------------------------------------------------------------------
{
	//NSLog(@"mufflonTaskFinished, trying to lock!\n");

	[resultProcessingLock lock];
	
	NSTask* task = (NSTask*) [aNotification object];

	//NSLog(@"mufflonTaskFinished, lock aqcuired, task: 0x%08x (%@)\n", task, task);

	[progressIndicator stopAnimation:self];	

	NSString* resultOutputPath = nil;
	BOOL batchResult = NO;
	
	int result = [task terminationStatus];
	if (task == mainMufflonTask)
	{
		[mainMufflonTask release];
		mainMufflonTask = nil;
		
		resultOutputPath = [outputPath copy];
	}
	else
	{
		[batchMufflonTasks removeObject:task];
		currentConversionsFinished++;
		[folderProgressIndicator setDoubleValue:currentConversionsFinished];

		resultOutputPath = [taskToOutputPath objectForKey:[NSNumber numberWithInt:(int)task]];
		batchResult = YES;
	}
	
	[convertButton setEnabled:YES];
	[nufliInputPathTextField setEnabled:YES];
	[nufliOutputPathTextField setEnabled:YES];
	[muifliInputPathTextField setEnabled:YES];
	[muifliOutputPathTextField  setEnabled:YES];
	[selectNuFliInputButton setEnabled:YES];
	[selectNuFliOutputButton setEnabled:YES];
	[selectMuiFliInputButton setEnabled:YES];
	[selectMuiFliOutputButton setEnabled:YES];
	[savePreConvertButton setEnabled:YES];
	
	if (result == 0)
	{
		if (mufflonTaskWasForPreview)
		{
			NSString* outputFilename = [resultOutputPath lastPathComponent];
			NSString* previewPath = [sTempPath stringByAppendingPathComponent:outputFilename];
			NSImage* previewImage = [[NSImage alloc] initByReferencingFile:previewPath];
			//NSLog(@"outputFilename: %@, previewPath: %@, previewImage: %@\n", outputFilename, previewPath, previewImage);

			NSImageView* imageView = [self getPreConvertImageView];
			[imageView setImage:previewImage];
			
			if (activeMode == MUIFLI_MODE)
			{
				[muifliLeftTabView selectTabViewItemWithIdentifier:@"2"];
				muifliPreConvertPath = previewPath;
			}
			else
			{
				[nufliLeftTabView selectTabViewItemWithIdentifier:@"2"];
				nufliPreConvertPath = previewPath;
			}
		}
		else
		{
			NSString* resultImagePath = nil;
			NSString* errorImagePath = nil;
			NSString* flickerImagePath = nil;
			
			NSString* resultDestinationPath = nil;
			NSString* errorDestinationPath = nil;
			NSString* flickerDestinationPath = nil;
			
			NSString* destinationPath = [resultOutputPath stringByDeletingLastPathComponent];
			
			if (activeMode == MUIFLI_MODE)
			{
				NSString* basePath = [sTempPath stringByAppendingPathComponent:baseName];
				resultImagePath = [basePath stringByAppendingString:@"_result.bmp"];
				errorImagePath = [basePath stringByAppendingString:@"_errormap.bmp"];
				flickerImagePath = [basePath stringByAppendingString:@"_flickermap.bmp"];

				//NSLog(@"base: '%@', result: '%@', error: '%@', flicker: '%@'\n", basePath, resultImagePath, errorImagePath, flickerImagePath);
				
				resultDestinationPath = [destinationPath stringByAppendingPathComponent:[baseName stringByAppendingString:@"_result.bmp"]];
				errorDestinationPath = [destinationPath stringByAppendingPathComponent:[baseName stringByAppendingString:@"_errormap.bmp"]];
				flickerDestinationPath = [destinationPath stringByAppendingPathComponent:[baseName stringByAppendingString:@"_flickermap.bmp"]];
			}
			else
			{
				NSString* basePath = [sTempPath stringByAppendingPathComponent:baseName];
				resultImagePath = [basePath stringByAppendingString:@"_result.bmp"];
				errorImagePath = [basePath stringByAppendingString:@"_errormap.bmp"];

				//NSLog(@"base: '%@', result: '%@', error: '%@'\n", basePath, resultImagePath, errorImagePath);

				resultDestinationPath = [destinationPath stringByAppendingPathComponent:[resultImagePath lastPathComponent]];
				errorDestinationPath = [destinationPath stringByAppendingPathComponent:[errorImagePath lastPathComponent]];
			}
			
			if (!batchResult)
			{
				NSImage* resultImage = [[NSImage alloc] initByReferencingFile:resultImagePath];
				NSImageView* imageView = [self getResultImageView];
				[imageView setImage:resultImage];
				
				NSImage* errorImage = [[NSImage alloc] initByReferencingFile:errorImagePath];
				imageView = [self getErrorImageView];
				[imageView setImage:errorImage];
				
				imageView = [self getFlickerImageView];
				if (flickerImagePath != nil)
				{
					NSImage* flickerImage = [[NSImage alloc] initByReferencingFile:flickerImagePath];
					[imageView setImage:flickerImage];
				}
				else
					[imageView setImage:nil];
			}
				
			NSFileManager* fileManager = [NSFileManager defaultManager];
			
			if (saveResults)
			{
				if ([fileManager fileExistsAtPath:resultImagePath])
				{
					[fileManager removeFileAtPath:resultDestinationPath handler:nil];
					[fileManager copyPath:resultImagePath toPath:resultDestinationPath handler:nil];
				}
				
				if ([fileManager fileExistsAtPath:errorImagePath])
				{
					[fileManager removeFileAtPath:errorDestinationPath handler:nil];
					[fileManager copyPath:errorImagePath toPath:errorDestinationPath handler:nil];
				}
				
				if (flickerImagePath != nil && [fileManager fileExistsAtPath:flickerImagePath])
				{
					[fileManager removeFileAtPath:flickerDestinationPath handler:nil];
					[fileManager copyPath:flickerImagePath toPath:flickerDestinationPath handler:nil];
				}
			}
			
			NSString* tempOutputPath = [sTempPath stringByAppendingPathComponent:[resultOutputPath lastPathComponent]];
			
			if ([fileManager fileExistsAtPath:tempOutputPath])
			{
				[fileManager removeFileAtPath:resultOutputPath handler:nil];
				[fileManager copyPath:tempOutputPath toPath:resultOutputPath handler:nil];
			}
		}
	}
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"An error occured!"
										 defaultButton:@"OK"
									   alternateButton:nil
										   otherButton:nil
							 informativeTextWithFormat:@"Check the Console.app output, or ask DeeKay..."];
		
		[alert runModal];
	}
	
	mufflonTaskFinished = YES;
	
	//NSLog(@"mufflonTaskFinished\n");
	
	[resultProcessingLock unlock];
	
	if (mufflonTaskWasForPreview && needsPreviewUpdate)
	{
		[self updatePreview];
		needsPreviewUpdate = NO;
	}
}



#pragma mark -
#pragma mark application delegate methods


// ----------------------------------------------------------------------------
- (BOOL) application:(NSApplication*)theApplication openFile:(NSString*)filename
// ----------------------------------------------------------------------------
{
	[self inputFileDropped:filename];
	return YES;
}



#pragma mark -
#pragma mark NSTabView delegate methods


// ----------------------------------------------------------------------------
- (void)tabView:(NSTabView *)tabView didSelectTabViewItem:(NSTabViewItem *)tabViewItem
// ----------------------------------------------------------------------------
{
	NSTabViewItem* item = [tabView selectedTabViewItem];
	[self setActiveMode:(EMode) [tabView indexOfTabViewItem:item]];
}


// ----------------------------------------------------------------------------
- (BOOL)tabView:(NSTabView *)tabView shouldSelectTabViewItem:(NSTabViewItem *)tabViewItem
// ----------------------------------------------------------------------------
{
	return mainMufflonTask == nil;
}

@end

#pragma mark -
#pragma mark NSImageView subclass


@implementation MufflonImageView

// ----------------------------------------------------------------------------
- (void) concludeDragOperation:(id <NSDraggingInfo>)sender
// ----------------------------------------------------------------------------
{
	NSString* path = [[[sender draggingPasteboard] propertyListForType:@"NSFilenamesPboardType"] objectAtIndex:0];
	AppDelegate* delegate = [NSApp delegate];
	[delegate inputFileDropped:path];
	
	//[super concludeDragOperation:sender];
}


@end


@implementation CustomProgressIndicator


static NSImage* walkerImage = nil;

// ----------------------------------------------------------------------------
- (void) drawRect:(NSRect)rect
// ----------------------------------------------------------------------------
{
	if (walkerImage == nil)
	{
		walkerImage = [[NSImage imageNamed:@"walker"] retain];
		isAnimating = NO;
		phase = 0.0f;
		alphaValue = 1.0f;
	}

	float width = [walkerImage size].width;
	float height = [walkerImage size].height;
	NSRect imageRect = NSMakeRect(0.0f, 0.0f, width, height);
	[walkerImage setFlipped:[self isFlipped]];
	
	[walkerImage drawInRect:rect fromRect:imageRect operation:NSCompositeSourceOver fraction:alphaValue];

	[super drawRect:rect];
}


// ----------------------------------------------------------------------------
- (void) animate
// ----------------------------------------------------------------------------
{
	if (isAnimating)
	{
		alphaValue = 0.5f + (0.4f * sinf(phase));
		phase += 0.4f;
	}
	else 
	{
		alphaValue = 1.0f;
	}
	
	[self setNeedsDisplay:YES];
}


// ----------------------------------------------------------------------------
- (void) startAnimation:(id)sender
// ----------------------------------------------------------------------------
{
	isAnimating = YES;
	phase = 0.0f;
	[self animate];
}


// ----------------------------------------------------------------------------
- (void) stopAnimation:(id)sender
// ----------------------------------------------------------------------------
{
	isAnimating = NO;
	[self animate];
}


@end



